/***********************************************************************
 * File: buffer_editor.c
 * 	(Extracted from
 *	   "$Id: buffer_editor.c,v 1.3 2009/05/30 22:53:17 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source file implements `buffer_editor' API-s.
 *
 * Copyright (C) 2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "buffer_editor.h"

/* type-definitions for this API-s... */
typedef enum substitute_e {
	  Substitute_matched_line
	, Substitute_constance
	, Substitute_sub_pattern
} Substitute_e;			/* for `id' member of `Substitute_st' */

typedef struct substitute_constance_t {
	char	*strp;	/* to the string */
	size_t	len;	/* length of it */
} Substitute_constance_t;	/* for `constance' of `Substitue_u' */

typedef union substitute_u {
	Substitute_constance_t	constance; /* `Substitute_matched_line'
					    * (`strp' must be NULL)
					    * or `Substitute_constance'
					    */
	size_t			pmatch_ix; /* `Substitute_sub_pattern'
					    */
} Substitute_u; 	/* for `value' of `Substitute_st' */

typedef struct substitute_st {
	Substitute_e		id;
	Substitute_u		value;
} Substitute_st;	/* for `parsed' member of `Substitute_t' */

typedef struct substitute_t {
	Substitute_st	*parsed;
	size_t		parsed_n;
	regmatch_t	*pmatch;
	size_t		nmatch;
} Substitute_t; /* parsed informations of `pattern2' */

typedef struct buffer_edit_regex_t {
	regex_t		preg;		/* for `pattern1' */
	Substitute_t	substitute;	/* for `pattern2' */
} Buffer_edit_regex_t; /* Finite-State-Machine (FSM) for editing buff.
			*/

typedef struct buffer_edit_line_nums_t {
	size_t	start, end;
	#define	Buffer_edit_line_nums_MAX_lineno ((size_t)-1)
} Buffer_edit_line_nums_t; /* range of target lines for editing */
#	define Buffer_edit_line_nums_N	2


/* release the `Buffer_edit_regex_t' typed structures */
static void destroy_parsed_pattern2(Substitute_t	*result)
{
	Substitute_st	*parsed = NULL;

	/* release all duplicated strings pointed from `parsed[]'... */
	for (parsed = result->parsed
	; result->parsed_n > 0
	; memset(parsed, (int)'\0', sizeof(parsed[0]))) {
		parsed	= result->parsed + (--(result->parsed_n));
		if ((parsed->id == Substitute_constance)
		&& parsed->value.constance.strp != NULL) {
			free(memset(parsed->value.constance.strp
				, (int)'\0'
				, (parsed->value.constance.len + 1)
				));
		}
	}

	/* release `parsed[]' array, if it was allocated. */
	if (result->parsed != NULL) {
		free(result->parsed);
	}
	result->parsed	= NULL;

	/* release `pmatch[]' array, if it was allocated. */
	if ((result->nmatch > 0) && result->pmatch != NULL) {
		free(memset(result->pmatch
			, (int)'\0'
			, (sizeof(result->pmatch[0]) * result->nmatch)
			));
	}
	result->nmatch = 0; result->pmatch = NULL;

	return;
}

/* extend the `parsed' array on the `Buffer_edit_regex_t' structures */
static Substitute_t *increment_parsed_array(Substitute_t *result
					, size_t	 increment_size
					)
{
	Substitute_st	*new_parsed;

	/* test: failed to allocate the extended `parsed[]' array? */
	if ((new_parsed = (Substitute_st *)
			realloc(result->parsed
				, (sizeof(result->parsed[0])
					* (result->parsed_n
						+ increment_size)
					)
				)) == NULL) {
		/* ERROR: insufficient memory. */
		destroy_parsed_pattern2(result)
		; errno = ENOMEM; return NULL; /* returns error */
	}

	/* initialize the extended elements of `parsed[]' array */
	result->parsed	= new_parsed
	; memset((result->parsed + result->parsed_n)
		, (int)'\0'
		, (sizeof(*(result->parsed)) * increment_size)
		);

	return	result; /* return normal-end */
}

/* extend the `pmatch' array on the `Buffer_edit_regex_t' structures */
static Substitute_t *increment_pmatch_array(Substitute_t	*result
					, size_t		nmatch
					)
{
	Substitute_st	*parsed;
	regmatch_t	*new_pmatch;

	/* test: failed to allocate the extended `pmatch[]' array? */
	parsed	= result->parsed + result->parsed_n;
	if ((new_pmatch = (regmatch_t *)
			realloc(result->pmatch
				, (sizeof(result->pmatch[0]) * nmatch)
				)) == NULL) {
		/* ERROR: insufficient memory. */
		destroy_parsed_pattern2(result)
		; errno = ENOMEM; return NULL; /* returns error */
	}

	/* set the number of `pmatch[]' array and address of extended */
	result->nmatch		= nmatch
	; result->pmatch	= new_pmatch
	;

	return	result;	/* return normal-end */
}

/* duplicate a string */
static char *dup_str(char *strp, size_t len)
{
	char	*result = (char *)
			malloc(len + 1); /* allocate new string */

	/* test: success to allocate the new string buffer? */
	if (result != NULL) {
		/* duplicate the `strp' string. */
		memcpy(result, strp, len); result[len] = '\0';
	}

	return	result;	/* if error occurred, return NULL.
			 * otherwise, return the address to new string.
			 */
}

/* parse the `pattern2' that string substitutes the REGEX matched. */
static Substitute_t *parse_pattern2(Substitute_t *result
				, char		 *pattern2
				, char		 pattern2_delimit
				)
{
	Substitute_st	*parsed;
	char		*elp, *ep, *sp;

	/* initialize `Substitute_t' structure */
	memset(result, (int)'\0', sizeof(*result));

	/* process `pattern2' string that is substitute-string... */
	for (sp = ep = pattern2; *ep != '\0'; ) {
		/* test: exist the backslash character? */
		if ((elp = strchr(ep, pattern2_delimit)) != NULL) {
			/* test: the next of backslash isn't digit? */
			if (isdigit(*(elp + 1)) == 0) {
				/* set `ep' to point to the digit */
				ep	= elp + 1;
				
				continue; /* process the rest... || */
			}

			/* test: failed to extend `parsed[]' array? */
			if ((result = increment_parsed_array(result, 2))
							== NULL) {
				break; /* returns error (=NULL). */
			}

			/* test: failed to duplicate constant string? */
			parsed	= result->parsed + result->parsed_n;
			if ((parsed->value.constance.strp
					= dup_str(sp
						, (size_t)(elp - sp)
						)) == NULL) {
				/* ERROR: insufficient memory */
				destroy_parsed_pattern2(result)
				; errno = ENOMEM; result = NULL
				;

				break; /* returns error (=NULL). */
			}

			/* set length of the duplicate strings and id
			 * of this `parsed[]' element.
			 */
			parsed->value.constance.len
				= strlen(parsed->value.constance.strp)
			; parsed->id	= Substitute_constance
			;

			/* process the backslash-digit sequence... */
			++(result->parsed_n); ++parsed; sp = elp + 1;

			/* test: the digit indicates too big number? */
			if (((parsed->value.pmatch_ix
					= (size_t)strtoul(sp, &sp, 10))
						== (size_t)ULONG_MAX)
			&& errno == ERANGE) {
				/* ERROR: out of range */
				destroy_parsed_pattern2(result)
				; result	= NULL
				;

				break; /* returns error (=NULL) */
			}

			/* test: next of the digit is
			 * `pattern2_delimit'?
			 */
			if (*sp == pattern2_delimit) {
				++sp;
			}

			/* test: the digit indicates zero? */
			if (parsed->value.pmatch_ix == 0UL) {
				/* set this `parsed[]' element to
				 * copy the matching line.
				 */
				parsed->id = Substitute_matched_line
				; parsed->value.constance.strp	= NULL
				; parsed->value.constance.len	= 0
				;
			}
			else {	/* set this `parsed[]' element to
				 * extract the sub-pattern string
				 */
				parsed->id = Substitute_sub_pattern;

				/* test: failed to extend `pmatch[]'
				 *	array, if the index of
				 *	sub-pattern string is out of
				 *	`pmatch[]' array size.
				 */
				if ((parsed->value.pmatch_ix + 1
						> result->nmatch)
				&& (result = increment_pmatch_array(
						result
						, (parsed->
							value.pmatch_ix
							+ 1)
						)) == NULL) {
					/* ERROR: insufficient memory */
					break; /* return error(=NULL) */
				}
			}

			/* increment number of `parsed[]' array */
			++(result->parsed_n);

			/* process the next characters in `pattern2' */
			ep = sp;
		}
		else {	/* exist no backslash character in `pattern2' */
			/* test: failed to extend `parsed[]' array? */
			if ((result = increment_parsed_array(result, 1))
							== NULL) {
				break; /* return error (=NULL). */
			}

			/* test: failed to duplicate this constant
			 *	string?
			 */
			parsed	= result->parsed + result->parsed_n;
			if ((parsed->value.constance.strp
						= dup_str(sp
							, strlen(sp)
							)) == NULL) {
				/* ERROR: insufficient memory */
				destroy_parsed_pattern2(result)
				; errno = ENOMEM; result = NULL
				;
				break; /* returns error (=NULL) */
			}

			/* set length of the duplicated string, set ID
			 * of this `parsed[]' element to copy the
			 * string, and increment number of `parsed[]'
			 */
			parsed->value.constance.len
				= strlen(parsed->value.constance.strp);

			parsed->id	= Substitute_constance
			; ++(result->parsed_n)
			;
			
			/* process the next character of `pattern2' */
			ep = sp = (ep + strlen(ep));
		}
	}

	return	result; /* returns result of parsing `pattern2' */
}



/* Predication means that `regexec(3)' returns matched.
 * ex. extracting the lines that matches `pattern1' from `src' buffer.
 */
static int matched_pred(int res_regexec)
{	return	(int)(res_regexec == 0);	}

/* Predication means that `regexec(3)' returns unmatched.
 * ex. deleting the lines that matches `pattern1' from `src' buffer.
 */
static int unmatched_pred(int res_regexec)
{	return	(int)(res_regexec != 0);	}

/* transfer the string sequence to the editing result buffer */
static int transfer_string(char		*sp
			, size_t	len
			, char		**dst
			, size_t	*dstsz
			, char		**dp
			, size_t	*leftsz
			)
{
	char	*new_dst;

	/* test: left size of `*dst' buffer is less than transfer
	 *	length?
	 */
	if (len >= (*leftsz - 1)) {
		/* extend the `*dst' buffer */
		size_t	new_dstsz = *dstsz - *leftsz + len + 1;

		/* test: failed to allocate the extended buffer? */
		if ((new_dst = (char *)realloc(*dst, new_dstsz))
							== NULL) {
			/* ERROR: insufficient memory */
			errno = ENOMEM; return !0; /* return error */
		}

		/* set the extended address, destiny address for coping,
		 * allocated size of `*dst' buffer and left size of it.
		 */
		*dst = new_dst; new_dst += (*dstsz - *leftsz)
		; *dstsz = new_dstsz; *leftsz = 1
		;
	}
	else {	/* left size of `*dst' buffer is thorough to transfer
		 * the string sequence
		 */
		new_dst = *dst + (*dstsz - *leftsz); *leftsz -= len;
	}

	/* transfer the string sequence to `*dst' buffer */
	memcpy(new_dst, sp, len); *dp = new_dst + len; **dp = '\0';

	return	0; /* return normal-end */
}

/* edit the matched line, according to `parsed[]' array settings */
static int edit_matched_line(Substitute_t	*substitute
			, char			*sp
			, size_t		len
			, char			**dst
			, size_t		*dstsz
			, char			**dp
			, size_t		*leftsz
			)
{
	regoff_t	rm_so;
	size_t		length, i;
	char		*srcp;
	int		result = 0;

	/* refer the each settings of `parsed[]' array... */
	for (i = 0; i < substitute->parsed_n; ++i) {
		/* test: which ID is this `parsed[]' element? */
		switch (substitute->parsed[i].id) {
		case Substitute_matched_line:
			/* copy the whole of matched line */
			length = ((sp[len - 1] == '\n')? len - 1: len)
			; srcp	= sp
			;
			break;

		case Substitute_constance:
			/* copy the constant string */
			srcp = substitute->parsed[i]
						.value.constance.strp
			; length = substitute->parsed[i]
						.value.constance.len
			;
			break;

		case Substitute_sub_pattern:
			/* copy the sub-pattern sequence */
			/* test: this sub-pattern is found? */
			if ((rm_so = substitute->pmatch[
						substitute->parsed[i]
							.value.pmatch_ix
					].rm_so) != (regoff_t)-1) {
				/* copy the found sub-pattern */
				srcp	= sp + rm_so
				; length = (size_t)(substitute->pmatch[
						substitute->parsed[i]
							.value.pmatch_ix
						].rm_eo - rm_so)
				;
			}
			else {	/* the sub-pattern isn't found,
				 * copy nothing
				 */
				srcp = NULL; length = 0;
			}
			break;

		default: /* FATAL: Oops! exist the fatal failures! */
			abort(); /* abort this process */
		}

		/* test: failed to transfer the string sequences? */
		if ((length > 0)
		&& transfer_string(srcp, length, dst, dstsz, dp, leftsz)
								!= 0) {
			result = !0; break; /* return error */
		}
	}

	return	result;	/* if it's normal-end, return zero.
			 * otherwise, return non-zero (error).
			 */
}

/* substitute the line that matched `pattern1' to `pattern2',
 * and copy all of the unmatched lines.
 */
static int substitute(char			*src
		, size_t			line_nums_n
		, Buffer_edit_line_nums_t	*line_nums
		, char				**dst
		, size_t			*dstsz
		, Buffer_edit_regex_t		*regex
		, int				substituted_only
		, int				general_substitute
		)
{
	size_t		i, lno, len, leftsz = *dstsz;
	char		*ep, *sp = src, *dp = *dst, sv_ch = '\0';
	int		result_regexec, is_target_line
			, substituted_something
			;

	/* process the `src' string... */
	for (lno = 0
	; (*sp != '\0')
		&& (line_nums_n == Buffer_edit_line_nums_N
			|| lno < line_nums[0].end)
	; ++lno) {
		substituted_something	= 0;
		do {
			/* test: exist the new-line character? */
			if ((ep = strchr(sp, '\n')) != NULL) {
				/* set the mark ('\0') of this line */
				sv_ch = *ep; *ep = '\0';
				len	= (size_t)((ep - sp) + 1);
			}
			else {	/* this line is the last one. */
				sv_ch = '\0';
				len	= strlen(sp);
			}

			/* test: this line is a target to edit? */
			if ((lno >= line_nums[0].start
				&& lno < line_nums[0].end)
			|| (line_nums_n == Buffer_edit_line_nums_N
				&& lno >= line_nums[1].start
				&& lno < line_nums[1].end)
			) {
				/* initialize the start offsets of
				 * `pmatch[]' array
				 */
				for (i = 0
				; i < regex->substitute.nmatch
				; ++i) {
					regex->substitute
						.pmatch[i].rm_so
					= regex->substitute
						.pmatch[i].rm_eo
					= (regoff_t)-1
					;
				}

				/* execute the FSM of REGEX using
				 * `regexec(3)'
				 */
				result_regexec
					= regexec((const regex_t *)
							(&(regex->preg))
						, (const char *)sp
						, regex->substitute
							.nmatch
						, regex->substitute
							.pmatch
						, 00000
						);

				is_target_line	= !0;
			}
			else {	/* this line isn't target to edit. */
				/* set to `unmatched' */
				result_regexec	= !0;
				is_target_line	= 0;
			}

			/* restore the new-line code */
			if (ep != NULL) {
				*ep	= sv_ch;
			}

			/* test: the line is matched to the regular
			 *	expression?
			 */
			if (matched_pred(result_regexec) != 0) {
				/* test: failed to copy the sub-string
				 *	that precedes the matched
				 *	sub-pattern.
				 */
				if (((regex->substitute.pmatch != NULL)
				&& regex->substitute.pmatch[0].rm_so
							>= 0)
				&& transfer_string(sp
						, (size_t)(regex->
							substitute
							.pmatch[0]
							.rm_so)
						, dst
						, dstsz
						, &dp
						, &leftsz
						) != 0) {
					return	!0; /* return error */
				}

				/* test: failed to edit this matched
				 *	line
				 */
				if (edit_matched_line(
						  &(regex->substitute)
						, sp
						, len
						, dst
						, dstsz
						, &dp
						, &leftsz
						) != 0) {
					return	!0; /* return error */
				}

				/* test: failed to append a new-line
				 *	code to the non-general
				 *	substituted strings.
				 */
				if ((((general_substitute == 0)
				&& regex->substitute.pmatch != NULL)
				&& (len - regex->substitute
						.pmatch[0].rm_eo) > 0)
				&& transfer_string(sp + regex->
							substitute
							.pmatch[0]
							.rm_eo
						, len - regex->
							substitute
							.pmatch[0]
							.rm_eo
						, dst
						, dstsz
						, &dp
						, &leftsz
						) != 0) {
					return	!0; /* return error */
				}

				/* go to the next line... */
				sp	+= (((general_substitute != 0)
						&& regex->substitute
							.pmatch != NULL
						)?
						((size_t)(regex->
							substitute
						  	.pmatch[0]
							.rm_eo
						  + ((sp[regex->
						  	substitute
							.pmatch[0]
							.rm_eo]
							== '\n')? 1: 0)
						))
						: len
					);
				substituted_something	= !0;
			}
			else if (len > 0) {
				/* this unmatched has some codes */
				/* test: failed to copy this unmatched
				 *	string? (if invoker needs to
				 *	copy unmatched strings)
				 */
				if (is_target_line != 0
				&& (substituted_only == 0
					|| (general_substitute != 0
					&& substituted_something != 0)
					)
				&& transfer_string(sp
						, len
						, dst
						, dstsz
						, &dp
						, &leftsz
						) != 0) {
					return	!0; /* return error */
				}

				/* go to the next line... */
				sp	+= len;

				/* break do-while loop,
				 * because this line was processed.
				 */
				break;
			}
		} while (general_substitute != 0);
	}

	return	0; /* return normal-end */
}

/* copy the extracted (or deleted) strings */
static int copy_the_strings(char			*src
			, size_t			line_nums_n
			, Buffer_edit_line_nums_t	*line_nums
			, char				*dst
			, regex_t			*preg
			, int			(*matching_predicate)(
							int res_regexec
						)
			)
{
	size_t		lno, sz;
	char		*ep, *sp = src, *dp = dst, sv_ch = '\0';

	/* process `src' string... */
	for (lno = 0
	; (*sp != '\0')
	&& (line_nums_n == Buffer_edit_line_nums_N
		|| lno < line_nums[0].end
		)
	; ++lno) {
		/* test: exist the new-line character? */
		if ((ep = strchr(sp, '\n')) != NULL) {
			/* set mark this line */
			sv_ch = *ep; *ep = '\0';
		}
		else {	/* this line is the last */
			sv_ch = '\0';
		}

		/* test: is this line a target to edit, and result of
		 *	execute the FSM of REGEX means that we must
		 *	process it?
		 */
		if (	(((lno >= line_nums[0].start)
				&& lno < line_nums[0].end
				)
			|| (((line_nums_n == Buffer_edit_line_nums_N)
				&& lno >= line_nums[1].start)
				&& lno < line_nums[1].end
				)
			)
		&& (*matching_predicate)(regexec((const regex_t *)preg
						, (const char *)sp
						, (size_t)0UL
						, (regmatch_t *)NULL
						, 00000
						)) != 0
		) {
			/* test: this line has some characters?
			 */
			if ((sz = ((ep != NULL)?
						(size_t)((ep - sp) + 1)
					:	strlen(sp)
					)) > 0) {
				/* restore the new-line */
				if (ep != NULL) {
					*ep	= sv_ch;
				}

				/* transfer this line */
				memcpy(dp, sp, sz); dp[sz] = '\0';
				
				/* process the next line... */
				dp += sz; sp += sz;
			}
		}
		else if (ep != NULL) {
			/* restore the new-line, and process the next
			 * line
			 */
			*ep = sv_ch; sp = ep + 1;
		}
		else {	/* process the next (it's the end of string)*/
			sp	+= strlen(sp);
		}
	}

	return	0;	/* return normal-end, always */
}

/* release the edited buffer that is allocated by `edit()'. */
static void release_edited_buffer(char	*edited_buffer
			, size_t	allocated_size
			)
{	free(memset(edited_buffer, (int)'\0', allocated_size));	}

/* edit `src' buffer, and return the address to editing results */
static char *edit(char			*src
		, size_t		start_line_no
		, size_t		end_line_no
		, Buffer_edit_command_e	cmd
		, Buffer_edit_regex_t	*regexp
		, int			release_FSM
		)
{
	Buffer_edit_line_nums_t	line_nums[Buffer_edit_line_nums_N];
	size_t			line_nums_n = 0, dstsz = 0;
	char			*dst = NULL;
	int			(*matching_predicate)(int res_regexec)
							= unmatched_pred
				, substituted_only = 0
				, result = !0
				;

	/* test: invalid arguments? */
	if ((((src == NULL)
	|| regexp == NULL)
	|| cmd < Buffer_edit_cmd_extract)
	|| cmd >= Buffer_edit_cmd_N) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return NULL;
	}

	/* allocate the pmatch array for `substitute' commands */
	if (((cmd == Buffer_edit_cmd_substitute
		|| cmd == Buffer_edit_cmd_substitute_only
		|| cmd == Buffer_edit_cmd_general_substitute
		|| cmd == Buffer_edit_cmd_general_substitute_only)
	&& regexp->substitute.nmatch == 0)
	&& regexp->substitute.pmatch == NULL) {
		if ((regexp->substitute.pmatch = (regmatch_t *)
			malloc(sizeof(regexp->substitute.pmatch[0])))
							== NULL) {
			/* ERROR: insufficient memory */
			if (release_FSM != 0) {
				/* release the compiled FSM */
				(void)buffer_edit_regfree(
							(void *)regexp
							);
			}
			errno = ENOMEM; return NULL;
		}
		memset(regexp->substitute.pmatch
			, (int)'\0'
			, sizeof(regexp->substitute.pmatch[0])
			);
		regexp->substitute.nmatch	= 1;
	}

	/* set the numbers as start and end of lines */
	if (start_line_no == 0) {
		start_line_no	= 1;
	}
	if (end_line_no == 0) {
		end_line_no	= (size_t)-1;
	}
	memset(line_nums, (int)'\0', sizeof(line_nums));
	if (start_line_no > end_line_no) {
		line_nums_n		= 2;
		line_nums[0].start	= 0;
		line_nums[0].end	= end_line_no;
		line_nums[1].start	= start_line_no - 1;
		line_nums[1].end   = Buffer_edit_line_nums_MAX_lineno;
	}
	else {
		line_nums_n		= 1;
		line_nums[0].start	= start_line_no - 1;
		line_nums[0].end	= end_line_no;
		line_nums[1].start = line_nums[1].end
				   = Buffer_edit_line_nums_MAX_lineno
				   ;
	}

	/* allocate memory for results of editing buffer */
	if ((dst = (char *)malloc((dstsz = strlen(src) + 1))) == NULL) {
		/* ERROR: insufficient memory */
		if (release_FSM != 0) {
			/* release the compiled FSM */
			(void)buffer_edit_regfree((void *)regexp);
		}
		errno = ENOMEM; return NULL;
	}
	memset(dst, (int)'\0', dstsz); /* initialize allocated memory */

	/* dispatch the editing commands... */
	switch (cmd) {
	case Buffer_edit_cmd_extract:
		matching_predicate	= matched_pred;
	case Buffer_edit_cmd_delete:
		result	= copy_the_strings(src
					, line_nums_n
					, line_nums
					, dst
					, &(regexp->preg)
					, matching_predicate
					);
		break;

	case Buffer_edit_cmd_substitute_only:
		substituted_only	= !0;
	case Buffer_edit_cmd_substitute:
		result	= substitute(src
				, line_nums_n
				, line_nums
				, &dst
				, &dstsz
				, regexp
				, substituted_only
				, 0
				);
		break;

	case Buffer_edit_cmd_general_substitute_only:
		substituted_only	= !0;
	case Buffer_edit_cmd_general_substitute:
		result	= substitute(src
				, line_nums_n
				, line_nums
				, &dst
				, &dstsz
				, regexp
				, substituted_only
				, !0
				);
		break;

	default: /* ERROR: unknown editing command (invalid argument) */
		errno	= EINVAL;
		break;
	}

	/* test: failed in editing? */
	if (((release_FSM != 0)
		&& buffer_edit_regfree((void *)regexp) != 0)
	|| result != 0) {
		/* ERROR: failed in editing buffer */
		release_edited_buffer(dst, dstsz)
		; dst = NULL; dstsz = 0
		;
	}

	/* trim the new-line code at the end of last line. */
	if (dst != NULL) {
		size_t	dst_len	= strlen(dst);
		char	*nlp;

		if (*(nlp = (dst + ((dst_len != 0)?	(dst_len - 1)
					:		dst_len
					))) == '\n') {
			*nlp	= '\0';
		}
	}

	return	dst; /* returns address to the edited buffer or NULL */
}



/* API of compiling FSM for editing buffer:
 * Arguments:
 *	o `pattern1' point to the regular expression
 *	o `cflags' indicates the `regcomp(3)' compiling flags that is
 *	 `REG_BASIC' or `logical or'-ed value of `REG_EXTENDED' and
 *	 `REG_ICASE'.
 *	o `pattern2' is pointing to substitute the REGEXP string.
 * 	  The substitute description is as like `sed(1)'.
 *	o `pattern2_delimit' specifies the character delimits the sub-
 *	 patterns. if it is nil (\0'), the default character is
 *	 backslash.
 *	o `regcomp_errcode' outputs the return code from `regcomp(3)',
 *	 when this API returns NULL, and the error is caused in the
 *	 regular expression libraries. If the regular expression
 *	 libraries occurred no errors, the output is `REG_OKAY'.
 *	  This value is able to convert the error message string,
 *	 using `regerror(3)'.
 *	o `regcomp_errbuf' outputs the error message of `regcomp(3)',
 *	 if `regcomp_errbuf' isn't NULL.
 *	o `regcomp_errbuf_size' indicates the size of `regcomp_errbuf'.
 *
 * Returns;
 *	o NULL: it means several errors occurred
 *	o non-NULL: it points to a compiled FSM for these API-s.
 */
Buffer_edit_FSM_t buffer_edit_regcomp(char	*pattern1
				, int		cflags
				, char		*pattern2
				, char		pattern2_delimit
				, int		*regcomp_errcode
				, char		*regcomp_errbuf
				, size_t	regcomp_errbuf_size
				)
{
	Buffer_edit_regex_t	*regexp;

	/* test: invalid arguments? */
	if ((((pattern1 == NULL)
	|| (cflags & REG_NOSUB) == REG_NOSUB)
	|| (cflags & REG_NEWLINE) == REG_NEWLINE)
	|| regcomp_errcode == NULL) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return (Buffer_edit_FSM_t)NULL;
	}

	/* if `pattern2_delimit' is nul (\0'), set it backslash. */
	if (pattern2_delimit == '\0') {
		pattern2_delimit = Buffer_edit_pattern_backslash[0];
	}

	/* test: failed to allocate `Buffer_edit_regex_t' area? */
	if ((regexp = (Buffer_edit_regex_t *)malloc(sizeof(*regexp)))
							== NULL) {
		/* ERROR: failed to allocate it. */
		errno = ENOMEM; return (Buffer_edit_FSM_t)NULL;
	}
	memset(regexp, (int)'\0', sizeof(*regexp)); /* init it */

	/* test: failed to parse `pattern2' (if `pattern2' isn't NULL)?
	 */
	if (((pattern2 != NULL) && pattern2[0] != '\0')
	&& parse_pattern2(&(regexp->substitute)
			, pattern2
			, pattern2_delimit
			) == NULL) {
		/* ERROR: occurred several errors in parsing `pattern2'
		 */
		/* release the compiled FSM */
		; free(memset((void *)regexp
				, (int)'\0'
				, sizeof(*regexp)
				))
		; regexp	= NULL
		;

		return	(Buffer_edit_FSM_t)NULL; /* return error */
	}

	/* test: invoker needs error messages as the result from
	 *	`regcomp(3)'?
	 */
	if ((regcomp_errbuf != NULL) && regcomp_errbuf_size > 0) {
		*regcomp_errbuf	= '\0'; /* init it */
	}

	/* test: failed in `regcomp(3)' for `pattern1'? */
	if ((*regcomp_errcode
		= regcomp((regex_t *)memset(&(regexp->preg)
						, (int)'\0'
						, sizeof(regexp->preg)
						)
			, (const char *)pattern1
			, cflags
			)) != REG_OKAY) {
		/* ERROR: failed in compiling regular expression */
		/* test: invoker needs the error message? */
		if (regcomp_errbuf != NULL) {
			/* get the error message of `regcomp(3)' */
			(void)regerror(*regcomp_errcode
					, &(regexp->preg)
					, regcomp_errbuf
					, regcomp_errbuf_size
					);
		}
		(void)buffer_edit_regfree(regexp)
		; regexp	= NULL
		; errno		= EINVAL
		;
		
		return (Buffer_edit_FSM_t)NULL;	/* return error */
	}

	/* returns address to the compiled FSM */
	return	(Buffer_edit_FSM_t)(regexp);
}


/* API of releasing the FSM that was compiled by `buffer_edit_regcomp()'
 * Arguments:
 *	o `regex' is pointing to the result of `buffer_edit_regcomp()'
 *	 (it is a compiled FSM for these API-s).
 *
 * Returns:
 *	o non-zero: it means occurred errors.
 *	o zero: it means normal-end.
 */
int buffer_edit_regfree(Buffer_edit_FSM_t	regex)
{
	Buffer_edit_regex_t	*regexp = (Buffer_edit_regex_t *)regex;

	/* test: invalid arguments? */
	if (regexp == NULL) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return !0;
	}

	/* release the compiled FSM */
	destroy_parsed_pattern2(&(regexp->substitute))
	; regfree(&(regexp->preg))
	; free(memset(regexp, (int)'\0', sizeof(*regexp)))
	;

	return	0;	/* returns normal-end */
}


/* API of editing buffer with the compiled FSM:
 * Arguments:
 *	o `src' is pointing to the source buffer.
 *	o `start_line_no' indicates number as start of lines.
 *	 if it is zero, this API assumes the number is one.
 *	o `end_line_no' indicates number as end of lines.
 *	 if it is zero, this API assumes the number is `(size_t)-1'.
 *	o `cmd' indicates the command of editing (extract, delete or
 *	  (non-general or general) substitute several lines in the `src'
 *	 buffer).
 *	o `regex' is pointing to the result of `buffer_edit_regcomp()'
 *	 (it is a compiled FSM for these API-s).
 *
 * Returns:
 *	o NULL: it means several errors occurred. `errno' indicates the
 *	 details of errors.
 *	o non-NULL: it is pointing to an allocated string that contains
 *	 the result of editing.
 */
char *buffer_edit_FSM(char			*src
			, size_t		start_line_no
			, size_t		end_line_no
			, Buffer_edit_command_e	cmd
			, Buffer_edit_FSM_t	regex
			)
{
	return	edit(	src
			, start_line_no
			, end_line_no
			, cmd
			, (Buffer_edit_regex_t *)regex
			, 0
			);
}


/* API of editing buffer:
 * Arguments:
 *	o `src' is pointing to the source buffer.
 *	o `start_line_no' indicates number as start of lines.
 *	 if it is zero, this API assumes the number is one.
 *	o `end_line_no' indicates number as end of lines.
 *	 if it is zero, this API assumes the number is `(size_t)-1'.
 *	o `cmd' indicates the command of editing (extract, delete or
 *	  (non-general or general) substitute several lines in the `src'
 *	 buffer).
 *	o `pattern1' is pointing to the REGEXP string.
 *	o `cflags' indicates the `regcomp(3)' compiling flags that is
 *	 `REG_BASIC' or `logical or'-ed value of `REG_EXTENDED' and
 *	 `REG_ICASE'.
 *	o `pattern2' is pointing to substitute the REGEXP string.
 * 	  The substitute description is like as `sed(1)'.
 *	o `pattern2_delimit' specifies the character delimits the sub-
 *	 patterns. if it is nil (\0'), the default character is
 *	 backslash.
 *	o `regcomp_errcode' outputs the return code from `regcomp(3)',
 *	 when this API returns NULL, and the error is caused in the
 *	 regular expression libraries. If the regular expression
 *	 libraries occurred no errors, the output is `REG_OKAY'.
 *	  This value is able to convert the error message string,
 *	 using `regerror(3)'.
 *	o `regcomp_errbuf' outputs the error message of `regcomp(3)',
 *	 if `regcomp_errbuf' isn't NULL.
 *	o `regcomp_errbuf_size' indicates the size of `regcomp_errbuf'.
 *	 
 * Returns:
 *	o NULL: it means several errors occurred. `errno' indicates the
 *	 details of errors.
 *	o non-NULL: it is pointing to the allocated memory that
 *	 contains the result of editing.
 */
char *buffer_edit(char			*src
		, size_t		start_line_no
		, size_t		end_line_no
		, Buffer_edit_command_e	cmd
		, char			*pattern1
		, int			cflags
		, char			*pattern2
		, char			pattern2_delimit
		, int			*regcomp_errcode
		, char			*regcomp_errbuf
		, size_t		regcomp_errbuf_size
		)
{
	Buffer_edit_regex_t	*regexp;

	/* compile the FSM as `pattern1' regular expression.
	 * (if `cmd' means extracting or deleting,
	 * set `pattern2' to NULL)
	 */
	if ((cmd == Buffer_edit_cmd_extract)
	|| cmd == Buffer_edit_cmd_delete) {
		pattern2	= NULL;
	}
	if ((regexp = (Buffer_edit_regex_t *)buffer_edit_regcomp(
						  pattern1
						, cflags
						, pattern2
						, pattern2_delimit
						, regcomp_errcode
						, regcomp_errbuf
						, regcomp_errbuf_size
						)) == NULL) {
		/* ERROR: failed to compile FSM of `pattern1' */
		return	NULL;
	}

	/* returns address to allocated memory,
	 * or NULL that means failed to edit it.
	 */
	return	edit(src, start_line_no, end_line_no, cmd, regexp, !0);
}


/* API to release the result of `buffer_edit()' and `buffer_edit_FSM()':
 * Arguments:
 *	o `edited_buffer' points to an allocated string that contains
 *	 the result of `buffer_edit()' (or `buffer_edit_FSM()').
 *
 * Returns:
 *	o non-zero: it means several errors occurred.
 *	o zero: it means normal-end.
 */
int buffer_edit_free(char	*edited_buffer)
{
	/* test: invalid arguments? */
	if (edited_buffer == NULL) {
		errno = EINVAL; return !0; /* ERROR: invalid args */
	}

	release_edited_buffer(edited_buffer
			, (strlen(edited_buffer) + 1)
			);	/* release the edited buffer */

	return	0;	/* returns normal-end */
}
