/***********************************************************************
 * File: conv_line_ctrl.c
 * 	(Extracted from
 *	   "$Id: conv_line_ctrl.c,v 1.3 2009/05/30 19:52:50 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements how to convert line control sequences in
 * the string.
 *
 * Copyright (C) 2006-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

/* include definitions for `conv_line_ctrl()' API */
#include "conv_line_ctrl.h"


/***********************************************************************
 * This API counts up the line
 *
 * arguments:
 *	src points to the source buffer.
 *	line_ctrl_sequence points to the sequence of line-control.
 *
 * returns:
 *	Non-zero indicates the number of lines.
 *	Zero means several errors occurred.
 *	(if this returns zero, errno indicates the details.
 *		EINVAL  (22)	: invalid arguments
 **********************************************************************/
size_t count_lines(char *src, char *line_ctrl_sequence)
{
	size_t	result, seq_len = 0;
	char	*lp;

	/* test: the arguments are illegal? */
	if (((src == NULL) || line_ctrl_sequence == NULL)
	|| (seq_len = strlen(line_ctrl_sequence)) == 0
	) {	errno = EINVAL; return 0; /* ERROR! invalid arguments */	}

	/* count up the line-control sequence... */
	for (result = 1; *src != '\0'; src = lp + seq_len, ++result) {
		if ((lp = strstr((const char *)src
						, (const char *)line_ctrl_sequence
						)) == NULL
		) {	break;	}
	}

	return	result;
}

/***********************************************************************
 * This API converts the line control sequences in a string
 *
 * arguments:
 *	style specifies the line-control styles.
 *	dst points to the destiny buffer.
 *	dst_sz indicates the size of destiny buffer.
 *	src points to the source buffer.
 *
 * returns:
 *	NULL means several errors are occurred.  Otherwise, this returns
 *	the address of destiny buffer.
 *	(if this returns NULL, errno indicates the details.
 *		EINVAL  (22)	: invalid arguments
 *		ENOBUFS (105)	: destiny buffer is insufficient.
 **********************************************************************/
char *conv_line_ctrl(Conv_line_ctrl_style_t	style
					, char					*dst
					, size_t				dst_sz
					, char					*src
					)
{
	char	*result = dst; /* initialize result of this API */

	/* test: invalid arguments? */
	if (((dst == NULL) || dst_sz == 0) || src == NULL) {
		errno = EINVAL; return NULL; /* ERROR: invalid args */
	}
	errno	= 0; /* initialize errno to `no errors' */

	/* test: which style is specified? */
#	define PROCESS_SRC	for (result = dst		\
				; (*src != '\0') && dst_sz > 1	\
				; --dst_sz			\
				)
	if (style == Conv_line_ctrl_to_Unix) {
		PROCESS_SRC {
			if (*src == '\r') {
				/* Dos or Mac line control sequence */
				src	+= ((*(src + 1) == '\n')? 2: 1);
				*dst++	= '\n';
			}
			else {
				/* other codes are transfered, as is */
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_Dos_to_Unix) {
		PROCESS_SRC {
			if ((*src == '\r') && *(src + 1) == '\n') {
				/* convert Dos line control, only */
				*dst++ = '\n'; src += 2;
			}
			else {
				/* other codes are transfered, as is */
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_Mac_to_Unix) {
		PROCESS_SRC {
			if ((*src == '\r') && *(src + 1) != '\n') {
				/* convert Mac line control, only */
				*dst++ = '\n'; ++src;
			}
			else {
				/* other codes are transfered, as is */
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_to_Dos) {
		PROCESS_SRC {
			if (*src == '\r') {
				if (dst_sz <= 2) { break; }

				if (*(src + 1) == '\n') {
					*dst++ = *src++; --dst_sz;
					*dst++ = *src++;
				}
				else {
					*dst++ = *src++; --dst_sz;
					*dst++ = '\n';
				}
			}
			else if (*src == '\n') {
				if (dst_sz <= 2) { break; }
				*dst++ = '\r'; --dst_sz;
				*dst++ = *src++;
			}
			else {
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_Unix_to_Dos) {
		PROCESS_SRC {
			if ((*src == '\r') && *(src + 1) == '\n') {
				if (dst_sz <= 2) { break; }

				*dst++ = *src++; --dst_sz;
				*dst++ = *src++;
			}
			else if (*src == '\n') {
				if (dst_sz <= 2) { break; }

				*dst++ = '\r'; --dst_sz;
				*dst++ = *src++;
			}
			else {
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_Mac_to_Dos) {
		PROCESS_SRC {
			if (*src == '\r') {
				if (dst_sz <= 2) { break; }

				if (*(src + 1) == '\n') {
					*dst++ = *src++; --dst_sz;
					*dst++ = *src++;
				}
				else {
					*dst++ = *src++; --dst_sz;
					*dst++ = '\n';
				}
			}
			else {
				*dst++	= *src++;
			}
		}
	}
	else if (style == Conv_line_ctrl_to_Mac) {
		PROCESS_SRC {
			if (*src == '\r') {
				*dst++	= *src;
				src	+= ((*(src + 1) == '\n')? 2: 1);
			}
			else {
				*dst++ = ((*src == '\n')? '\r': *src);
				++src;
			}
		}
	}
	else if (style == Conv_line_ctrl_Unix_to_Mac) {
		PROCESS_SRC {
			if (*src == '\r') {
				if (*(src + 1) == '\n') {
					if (dst_sz <= 2) { break; }
					*dst++	= *src++; --dst_sz;
				}
				*dst++	= *src++;
			}
			else {
				*dst++ = ((*src == '\n')? '\r': *src);
				++src;
			}
		}
	}
	else if (style == Conv_line_ctrl_Dos_to_Mac) {
		PROCESS_SRC {
			if (*src == '\r') {
				if (*(src + 1) == '\n') {
					*dst++	= *src; src += 2;
				}
				else {
					*dst++	= *src++;
				}
			}
			else {
				*dst++ = *src++;
			}
		}
	}
	else {
		result	= NULL;
	}

	if (result != NULL) {
		*dst	= '\0'; /* put the terminal code of string */
	}

	if (*src != '\0') {
		errno	= ENOBUFS;
	}

	return	(errno != 0)? NULL: result; /* NULL means occurred
										 * several errors.
										 * Otherwise, this returns
										 * address of the
										 * destination string
										 */
#	undef PROCESS_SRC
}
