/***********************************************************************
 * File: simple_hashing.h
 * 	(Extracted from
 *	   "$Id: simple_hashing.h,v 1.3 2009/06/05 19:45:57 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C header defines the simple hashing API and data structures.
 *
 * Copyright (C) 2005-2009 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#if !defined(_SIMPLE_HASHING_H_) /* { */
#define	_SIMPLE_HASHING_H_
#ifdef __cplusplus
extern "C" {
#endif

#include <unistd.h>
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <limits.h>

/*
 * Hint for datum counts of 'elements' array pointed from 'HashTable_t'
 * for the first time allocation array length, the extension length and
 * the maximum datum counts.
 * 
 * This hints are adjusted and set to several 'Hashtable_t' members by
 * 'create_hash_value()'.  And then, the members are refereed from
 * '_insert_hash_table()' that try to create/extend the 'elements' array
 * pointed from 'Hashtable_t'.
 */
typedef struct hashtable_extend_hint_t {
	ssize_t	max_insert_cnt;	/* maximum count of datum */
#	define	HASHTABLE_SYS_MAX_INSERT_CNT	(0L)
	ssize_t	elems_base_cnt;	/* datum count of the first allocation
							 * "elements" array
							 */
#	define	HASHTABLE_RECOMMENDED_ELEMS_BASE_CNT (0L)
	ssize_t	elems_ext_cnt;	/* datum count of the extension datum counts for
							 * "elements" array
							 */
#	define HASHTABLE_NEVER_EXTEND (0L)
#	define HASHTABLE_RECOMMENDED_ELEMS_EXT_CNT (-1L)
} Hashtable_extend_hint_t;

/*
 * An element of the hash table that is pointing to 'elements' array
 * of the datum elements.
 * A 'Hashtable_elems_t' can hold LONG_MAX datums in 'elements' array.
 */
typedef struct hashtable_elems_t {
	ssize_t	duplicated_cnt;	/* count of duplicated hash-value elements */
	ssize_t	elements_len;	/* length of `elements' */
	void	**elements; /* array of elements that the same hash-value */
} Hashtable_elems_t;

#define IS_HASHTABLE_ELEMS_ELEMENTS_FULL(HASHTABLE_ELEMS)				\
					((HASHTABLE_ELEMS)->elements_len == (ssize_t)LONG_MAX)
#define IS_HASHTABLE_ELEMS_ELEMENTS_ALLOCATED(HASHTABLE_ELEMS)			\
									((HASHTABLE_ELEMS)->elements_len > 0L)

/*
 * The hash table structure holds 'hashtable_elems_t' array.
 * A 'Hashtable_t' can hold ULONG_MAX datum in the whole 'tables' array.
 */
typedef struct hashtable_t {
	ssize_t	(*make_hash_value)(char **keys, size_t n_keys, ssize_t depth);
								/* user specified hash-function */
	int		(*cmp_keys_element)(char **keys, size_t n_keys, void *elem);
								/* user specified compare-function */
	ssize_t	elem_cnt_per_page;	/* the datum counts in a page */
#	define	HASHTABLE_SYSTEM_PAGESIZE_DEFAULT (4L * 1024L)
	ssize_t	depth;				/* max number of element */
	ssize_t	used_cnt; 			/* count of used `tables' elements */
	ssize_t max_insert_cnt;		/* maximum counts of inserted datum */
	ssize_t alloc_base_cnt;		/* the first time allocation datum count */
	ssize_t	alloc_ext_cnt;		/* the extension datum count */
	ssize_t	inserted_cnt;		/* count of inserted `elements' */
	struct hashtable_elems_t	tables[1]; /* hash-table allocated area */
} Hashtable_t;

#define IS_HASHTABLE_TABLES_FULL(HASHTABLE)								\
				((HASHTABLE)->inserted_cnt == (HASHTABLE)->max_insert_cnt)


/*
 * This API creates the new hash table structure.
 * Invoker must specify the depth of hash table and the function that
 * compares keys and an element.
 * Invoker is accepted to specify a hash value maker function
 * ('make_hash_value' arg). If invoker doesn't set `make_hash_value',
 * the function is set to the default 'simple_hash_value()'.
 *
 * If 'hint' isn't NULL, the parameters adjusted and set to the several
 * members of 'Hashtable_t'.  Otherwise, this API sets the defaults to
 * those.  The members are referred by the private method
 * '_insert_hash_table()' used by 'insert_hashtable()' and
 * 'replace_hash_table()' API-s.
 *
 * If several errors occurred, this returns NULL. The occurred error
 * detail is indexed by 'errno'.
 *	'errno': means...
 *	EINVAL : illegal arguments
 *	ENOMEM : insufficient memories for 'malloc()'
 *
 * Arguments means...
 *	depth				- size of the creating hash table.
 *	make_hash_value		- pointer to function that makes the hash value from
 *					  	  key strings.  If it is set NULL, the hash table
 *					  	  takes the default private method
 *					  	  'simple_hash_value()'
 *	cmp_keys_element	- pointer to function that comparing key strings
 *						  and the datum.  it must be set user defined
 *						  function address.
 *	hint				- pointer to the hint parameters that specifies the
 *						  datum counts for allocation 'elements' array and
 *						  limitation the insert datum.
 */
extern Hashtable_t *create_hash_table(ssize_t depth
									, ssize_t (*make_hash_value)(
															  char **keys
															, size_t n_keys
															, ssize_t depth
															)
									, int (*cmp_keys_elements)(
															  char **keys
															, size_t n_keys
															, void *elem
															)
									, Hashtable_extend_hint_t *hint
									);
#define ILLEGAL_HASH_VAL (-1L)


/*
 * This API returns size of the allocated ares of the hash table.
 * This returns size that is equal or greater than sizeof(Hashtable_t).
 *
 * If the several errors occurred, this returns zero and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments (maybe, hash table isn't sane.)
 *
 * Arguments means...
 *	table	- pointer to the hash table
 */
extern size_t get_allocated_size_hash_table(Hashtable_t *table);


/*
 * This API returns the counts of the datum elements those were inserted
 * the hash table, yet.
 *
 * If the several errors occurred, this returns ULONG_MAX and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments (maybe, hash table isn't sane.)
 *
 * Arguments means...
 *	table	- pointer to the hash table
 */
extern size_t get_inserted_cnt_hash_table(Hashtable_t *table);


/*
 * This API returns the counts of used elements of hash table 'tables'.
 *
 * If the several error occurred, this returns LONG_MIN and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments
 *
 * Arguments means...
 *	table	- pointer to the hash table
 */
extern ssize_t get_used_cnt_hash_table(Hashtable_t *table);


/*
 * This API inserts the datum into 'elements' array of the hash table.
 * This returns the hash table that was inserted the datum.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	errno	: means...
 *	EINVAl	: illegal arguments
 *	ERANGE	: 'make_hash_value' function returns negative integer.
 *	ENOENT	: full of the inserted datum
 *	ENOMEM	: insufficient memories for 'malloc()', or this hash table
 *			  never extend the 'elements' array that was allocated once.
 *	ESRCH	: the same keys datum was inserted, yet.
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	keys	- pointer to pointer to key strings
 *	n_keys	- size of 'keys' pointer array
 *	elem	- pointer to the inserting datum
 */
extern Hashtable_t *insert_hash_table(Hashtable_t	*table
									, char			**keys
									, size_t		n_keys
									, void			*elem
									);


/*
 * This API replace the datum in 'elements' array of the hash table.
 * If the keys matching datum wasn't inserted yet, this inserts the datum
 * into 'elements' array.
 * This returns the hash table that was replaced/inserted the datum.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	errno	: means...
 *	EINVAl	: illegal arguments
 *	ERANGE	: 'make_hash_value' function returns negative integer.
 *	ENOMEM	: insufficient memories for 'malloc()', or this hash table
 *			  never extend the 'elements' array that was allocated once.
 *	ENOMEM	: insufficient memories for 'malloc()'
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	keys	- pointer to pointer to key strings
 *	n_keys	- size of 'keys' pointer array
 *	elem	- pointer to the replacing datum
 */
extern Hashtable_t *replace_hash_table(Hashtable_t	*table
										, char		**keys
										, size_t	n_keys
										, void		*elem
										);


/*
 * This API returns the counts of datum in the same 'elements' array.
 *
 * If several errors occurred, this return LONG_MIN and set 'errno'.
 *	errno	: means...
 *	EINVAl	: illegal arguments
 *	ERANGE	: 'make_hash_value' function returns negative integer.
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	keys	- pointer to pointer to key strings
 *	n_keys	- size of 'keys' pointer array
 */
extern ssize_t get_duplicated_cnt_hash_table(Hashtable_t	*table
											, char			**keys
											, size_t		n_keys
											);


/*
 * This API returns the datum in the hash table that is matched
 * the keys.
 *
 * If several errors occurred, this returns NULL and set "errno".
 *	errno	: means
 *	EINVAL	: illegal arguments
 *	ERANGE	: 'make_hash_value' function returns negative integer.
 *	ENOENT	: nothing is matched keys.
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	keys	- pointer to pointer to the key strings
 *	n_keys	- size of 'keys' pointer array
 */
extern void *get_elem_hash_table(Hashtable_t	*table
								, char			**keys
								, size_t		n_keys
								);


/*
 * This API deletes the datum from the hash table 'elements' array.
 * This returns the hash table that is deleted the datum.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments
 *	ERANGE	: 'make_hash_value' function returns negative integer.
 *	ENOENT	: unfounded the datum in the hash table
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	keys	- pointer to pointer to the key strings
 *	n_keys	- size of 'keys' pointer array
 *	elem	- pointer to the deleting datum
 */
extern Hashtable_t *delete_hash_table(Hashtable_t	*table
									, char			**keys
									, size_t		n_keys
									, void			*elem
									);


/*
 * This API provides the cursor mechanism for the hash table.
 * The cursor will be used as iterator for the hash table.
 * This returns the current datum address of 'elements' in the 'tables'
 * member of 'table'.
 * When '*cursor' equals NULL, this allocates memory of the new cursor,
 * set the current position to the first datum in the hash table and
 * returns the first datum address.
 * When 'direct' equals 'HASHTABLE_CURSOR_DIR_DESTROY' this returns
 * 'table' address.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments
 *	ENOMEM	: insufficient memory
 *
 * Arguments means...
 *	table	- pointer to the hash table
 *	cursor	- address to output the created cursor, or pointer to the cursor
 *	direct	- switch to action of this API.
 *			direct							: action of API
 *			HASHTABLE_CURSOR_DIR_CURRENT	: return the current datum
 *			HASHTABLE_CURSOR_DIR_PREVIOUS	: return the previous datum
 *											  or NULL
 *			HASHTABLE_CURSOR_DIR_NEXT		: return the next datum or NULL
 *			HASHTABLE_CURSOR_DIR_DESTROY	: release resources of the
 *											  cursor
 */
typedef struct hashtable_cursor_t {
	Hashtable_t	*table;
	ssize_t		tables_idx;
	ssize_t		walked_tables_cnt;
	ssize_t		elements_idx;
} Hashtable_cursor_t;

typedef enum hashtable_cursor_dir_t {
	  HASHTABLE_CURSOR_DIR_CURRENT	= 0
	, HASHTABLE_CURSOR_DIR_PREVIOUS
	, HASHTABLE_CURSOR_DIR_NEXT
	, HASHTABLE_CURSOR_DIR_DESTROY
} Hashtable_cursor_dir_t;

extern void *get_elem_by_hashtable_cursor(Hashtable_t				*table
										, Hashtable_cursor_t		**cursor
										, Hashtable_cursor_dir_t	direct
										);


/*
 * This API release the allocated areas around the hash table.
 * This returns zero, when the hash table resources are released,
 * completely.
 *
 * If several errors occurred, this returns -1 and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments
 *
 * Arguments means...
 *	table	- pointer to the hash table
 */
extern int destroy_hash_table(Hashtable_t *table);


/*
 * This API provides the binary search method.
 * This invokes user specified compare function 'cmp_keys_elements'.
 * If the datum is found, this return non-zero.
 * Otherwise, this returns zero (if this gets illegal arguments.
 * 'errno" is set EINVAL).
 * This outputs the index of the found/insert datum in the array.
 *
 * If several errors occurred, this returns zero and set 'errno'.
 *	errno	: means...
 *	EINVAL	: illegal arguments
 *
 * Arguments means...
 *	cmp_keys_elements	- pointer to function that comparing key strings
 *						  and the datum in sorted array
 *	array				- pointer to array of sorted datum
 *	n_array				- size of 'array' sorted datum array
 *	keys				- pointer to key strings array
 *	n_keys				- size of 'keys' string array
 *	pos					- address to output the index of found/inserting
 *						  datum position in the 'array'
 *	pos_cmp_result		- address to output the result of comparing 'keys'
 *						  strings and datum is indexed from '*pos'
 */
extern int bsearch_array(int		(*cmp_keys_elements)(char		**keys
														, size_t	n_keys
														, void		*elem
														)
						, void		**array
						, ssize_t	n_array
						, char		**keys
						, size_t	n_keys
						, ssize_t	*pos
						, int		*pos_cmp_result
						);


/*
 * The followings implements the unit test codes for this API-s.
 * If 'UnitTest' macro is defined, this unit test is enable.
 */
#if defined(UnitTest) /* { Unit Test program */

/*
 * This unit test method makes the hash value from keys,
 * using 'simple_hash_value()' that is defined above.
 *
 * Arguments means...
 *	keys	- pointer to the key string array
 *	n_keys	- size of 'keys' string array
 *	depth	- size of the hash value
 */
extern ssize_t make_hashval(char **keys, size_t n_keys, ssize_t depth);

#endif /* } defined(UnitTest) Unit Test program */

#ifdef __cplusplus
}
#endif
#endif /* defined(_SIMPLE_HASHING_H_) } */
