/***********************************************************************
 * File: sort_linked_list.h
 * 	(Extracted from
 *	   "$Id: sort_linked_list.h,v 1.2 2009/05/30 13:40:42 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This is the C header file that defines the 'sorting linked list' API
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#if !defined(_SORT_LINKED_LIST_H_) /* { */
#define _SORT_LINKED_LIST_H_
#ifdef __cplusplus
extern "C" {
#endif

#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <memory.h>
#include <errno.h>

/*
 * Abstracts;
 *	The "sort_linked_list()" API implements the sorting utility for
 * unidirectional and bidirectional linked lists, using "qsort()".
 * If the linked list was sorted completely, this returns the address of
 * the first node of the sorted.
 *
 * Arguments:
 *	  'list' points to the linked list.
 *	  'nodes_n' indicates the number of nodes in the linked list.
 *	If the negative integer is specified, it means the number is
 *	unknown.
 *	  'next' is the address of function that must return the address
 *	of the next node.
 *	  'prev (OPTIONAL)' is the address of function that must return
 *	the address of the previous node.
 *	  'get_datum_addr (OPTIONAL)' is the address of function that
 *	must return the address of the datum that must be pointed from
 *	the node.
 *	  'get_datum_fld_addr_on_node (OPTIONAL)' is the address of
 *	function that must return the address of the member of node,
 *	the member must be pointing to the datum.
 *	  'cmp_nodes_via_ptr2ptr' is the address of  function that
 *	compares two nodes or datums those are pointed by the
 *	pointer-to-pointer.
 *	  'set_next' is the address of function that sets the next node
 *	address to the node's 'next' member.
 *	  'set_prev (OPTIONAL)' is the address of function that sets the
 *	previous node address to the node's 'previous' member.
 *	  'index (OPTIONAL)' is the pointer to array of the nodes that
 *	is ordered as the nodes have been sorted.  If it is set NULL,
 *	this API will allocate the memory for array and output the
 *	allocated address to this.  The allocated are must be
 *	'free()'-ed by invokers.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	'errno'	: means...
 *	EINVAL	: illegal arguments
 *	ENOMEM	: insufficient memory
 */
void *sort_linked_list(void		*list
					, ssize_t	nodes_n
#						define	UNKNOWN_NUMBER_OF_SORTING_NODES (-1)
					, void		*(*next)(const void *list)
					, void		*(*prev)(const void *list)
					, void		*(*get_datum_addr)(const void *node)
					, void		*(*get_datum_fld_addr_on_node)(
															const void *node
															)
					, int		(*cmp_nodes_via_ptr2ptr)(const void *node1
														, const void *node2
														)
					, void		*(*set_next)(void *src_node
											, const void *next_node
											)
					, void		*(*set_prev)(void *src_node
											, const void *prev_node
											)
					, void		***vector
					);

/*
 * Abstracts:
 *	  This macro sorts the datum that are pointing from the nodes
 *	in the linked list.
 *
 * Arguments:
 *	  'L' points to the linked list.
 *	  'N' indicates the number of nodes in the linked list.  If the
 *	negative integer is specified, it means the number is unknown.
 *	  'NX' is the address of function that must return the address
 *	of the next node.
 *	  'PV (OPTIONAL)' is the address of function that must return
 *	the address of the previous node.
 *	  'GTDAD' is the address of function that must return the
 *	address of the datum that must be pointed from
 *	the node.
 *	  'GTFAD' is the address of function that must return the
 *	address of the member of node, the member must be pointing to
 *	the datum.
 *	  'CMPP2P' is the address of  function that compares two nodes
 *	or datums those are pointed by the pointer-to-pointer.
 *	  'V' is the pointer to array of the nodes that is ordered as
 *	the nodes have been sorted.
 *	  'SETPV (OPTIONAL)' is the address of function that sets the
 *	previous node address to the node's 'previous' member in the
 *	bidirectional linked list.
 *	  'IDX (OPTIONAL)' is the pointer to array of the nodes that
 *	is ordered as the nodes have been sorted.  If it is set NULL,
 *	this API will allocate the memory for array and output the
 *	allocated address to this.  The allocated are must be
 *	'free()'-ed by invokers.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	'errno'	: means...
 *	EINVAL	: illegal arguments
 *	ENOMEM	: insufficient memory
 */
#define sort_datum_in_linked_list(L,N,NX,PV,GTDAD,GTFAD,CMP,SETPV,IDX)	\
			(sort_linked_list((L), (N), (NX), (PV), (GTDAD), (GTFAD)	\
										, (CMP), NULL, (SETPV), (IDX)	\
								))

/*
 * Abstracts:
 *	This macro sorts the nodes and reconfigures the linked list.
 *
 * Arguments:
 *	  'L' points to the linked list.
 *	  'N' indicates the number of nodes in the linked list.  If the
 *	negative integer is specified, it means the number is unknown.
 *	  'NX' is the address of function that must return the address
 *	of the next node.
 *	  'PV (OPTIONAL)' is the address of function that must return
 *	the address of the previous node.
 *	  'CMPP2P' is the address of  function that compares two nodes
 *	or datums those are pointed by the pointer-to-pointer.
 *	  'SETNX' is the address of function that sets the next node
 *	address to the node's 'next' member.
 *	  'SETPV (OPTIONAL)' is the address of function that sets the
 *	previous node address to the node's 'previous' member in the
 *	bidirectional linked list.
 *	  'IDX (OPTIONAL)' is the pointer to array of the nodes that
 *	is ordered as the nodes have been sorted.  If it is set NULL,
 *	this API will allocate the memory for array and output the
 *	allocated address to this.  The allocated are must be
 *	'free()'-ed by invokers.
 *
 * If several errors occurred, this returns NULL and set 'errno'.
 *	'errno'	: means...
 *	EINVAL	: illegal arguments
 *	ENOMEM	: insufficient memory
 */
#define sort_nodes_in_linked_list(L,N,NX,PV,CMPP2P,SETNX,SETPV,IDX)		\
					(sort_linked_list((L), (N), (NX), (PV), NULL, NULL	\
									, (CMPP2P), (SETNX), (SETPV), (IDX)	\
									))

#ifdef __cplusplus
}
#endif
#endif /* !defined(_SORT_LINKED_LIST_H_) } */
